
/*
 * original kernel code by:	Eddie C. Dost
 *				David S. Miller
 *				Andre Hedrick
 */

#include <stdio.h>
#include <unistd.h>
#include <sys/io.h>

#include "atapci.h"

/*
 * CMD64x specific registers definition.
 */

#define CFR		0x50
#define   CFR_INTR_CH0		0x02
#define CNTRL		0x51
#define	  CNTRL_DIS_RA0		0x40
#define   CNTRL_DIS_RA1		0x80
#define	  CNTRL_ENA_2ND		0x08

#define	CMDTIM		0x52
#define	ARTTIM0		0x53
#define	DRWTIM0		0x54
#define ARTTIM1 	0x55
#define DRWTIM1		0x56
#define ARTTIM23	0x57
#define   ARTTIM23_DIS_RA2	0x04
#define   ARTTIM23_DIS_RA3	0x08
#define   ARTTIM23_INTR_CH1	0x10
#define ARTTIM2		0x57
#define ARTTIM3		0x57
#define DRWTIM23	0x58
#define DRWTIM2		0x58
#define BRST		0x59
#define DRWTIM3		0x5b

#define BMIDECR0	0x70
#define MRDMODE		0x71
#define BMIDESR0	0x72
#define UDIDETCR0	0x73
#define DTPR0		0x74
#define BMIDECR1	0x78
#define BMIDECSR	0x79
#define BMIDESR1	0x7A
#define UDIDETCR1	0x7B
#define DTPR1		0x7C

static struct cmd64x_dev {
	u16 device;
	char* name;
} cmd64x_devs[] = {
	{ PCI_DEVICE_ID_CMD_649, "649" },
	{ PCI_DEVICE_ID_CMD_648, "648" },
	{ PCI_DEVICE_ID_CMD_646, "646" },
	{ PCI_DEVICE_ID_CMD_643, "643" },
};

int cmd64x_info(char *buffer, struct pci_dev *dev)
{
	char *p = buffer;
	u8 reg53, reg54, reg55, reg56;	/* primary */
	u8 reg57, reg58, reg5b;		/* secondary */
	u8 reg72, reg73;		/* primary */
	u8 reg7a, reg7b;		/* secondary */
	u8 reg50, reg71;		/* extra */
	unsigned int i;

	for (i = 0; i < ARRAY_SIZE(cmd64x_devs); i++)
		if (dev->device_id == cmd64x_devs[i].device)
			break;

	p += sprintf(p, "\n\t\t\t\t      CMD%s Chipset.\n",
			cmd64x_devs[i].name);

	pci_read_config_byte(dev, CFR,       &reg50);
	pci_read_config_byte(dev, ARTTIM0,   &reg53);
	pci_read_config_byte(dev, DRWTIM0,   &reg54);
	pci_read_config_byte(dev, ARTTIM1,   &reg55);
	pci_read_config_byte(dev, DRWTIM1,   &reg56);
	pci_read_config_byte(dev, ARTTIM2,   &reg57);
	pci_read_config_byte(dev, DRWTIM2,   &reg58);
	pci_read_config_byte(dev, DRWTIM3,   &reg5b);
	pci_read_config_byte(dev, MRDMODE,   &reg71);
	pci_read_config_byte(dev, BMIDESR0,  &reg72);
	pci_read_config_byte(dev, UDIDETCR0, &reg73);
	pci_read_config_byte(dev, BMIDESR1,  &reg7a);
	pci_read_config_byte(dev, UDIDETCR1, &reg7b);

	p += sprintf(p, "--------------- Primary Channel ---------------- Secondary Channel -------------\n");
	p += sprintf(p, "                %sabled                         %sabled\n",
		(reg72&0x80)?"dis":" en",(reg7a&0x80)?"dis":" en");
	p += sprintf(p, "--------------- drive0 --------- drive1 -------- drive0 ---------- drive1 ------\n");
	p += sprintf(p, "DMA enabled:    %s              %s             %s               %s\n",
		(reg72&0x20)?"yes":"no ",(reg72&0x40)?"yes":"no ",(reg7a&0x20)?"yes":"no ",(reg7a&0x40)?"yes":"no ");
	p += sprintf(p, "DMA Mode:       %s(%s)          %s(%s)         %s(%s)           %s(%s)\n",
		(reg72&0x20)?((reg73&0x01)?"UDMA":" DMA"):" PIO",
		(reg72&0x20)?(  ((reg73&0x30)==0x30)?(((reg73&0x35)==0x35)?"3":"0"):
				((reg73&0x20)==0x20)?(((reg73&0x25)==0x25)?"3":"1"):
				((reg73&0x10)==0x10)?(((reg73&0x15)==0x15)?"4":"2"):
				((reg73&0x00)==0x00)?(((reg73&0x05)==0x05)?"5":"2"):"X"):"?",
		(reg72&0x40)?((reg73&0x02)?"UDMA":" DMA"):" PIO",
		(reg72&0x40)?(	((reg73&0xC0)==0xC0)?(((reg73&0xC5)==0xC5)?"3":"0"):
				((reg73&0x80)==0x80)?(((reg73&0x85)==0x85)?"3":"1"):
				((reg73&0x40)==0x40)?(((reg73&0x4A)==0x4A)?"4":"2"):
				((reg73&0x00)==0x00)?(((reg73&0x0A)==0x0A)?"5":"2"):"X"):"?",
		(reg7a&0x20)?((reg7b&0x01)?"UDMA":" DMA"):" PIO",
		(reg7a&0x20)?(	((reg7b&0x30)==0x30)?(((reg7b&0x35)==0x35)?"3":"0"):
				((reg7b&0x20)==0x20)?(((reg7b&0x25)==0x25)?"3":"1"):
				((reg7b&0x10)==0x10)?(((reg7b&0x15)==0x15)?"4":"2"):
				((reg7b&0x00)==0x00)?(((reg7b&0x05)==0x05)?"5":"2"):"X"):"?",
		(reg7a&0x40)?((reg7b&0x02)?"UDMA":" DMA"):" PIO",
		(reg7a&0x40)?(	((reg7b&0xC0)==0xC0)?(((reg7b&0xC5)==0xC5)?"3":"0"):
				((reg7b&0x80)==0x80)?(((reg7b&0x85)==0x85)?"3":"1"):
				((reg7b&0x40)==0x40)?(((reg7b&0x4A)==0x4A)?"4":"2"):
				((reg7b&0x00)==0x00)?(((reg7b&0x0A)==0x0A)?"5":"2"):"X"):"?" );
	p += sprintf(p, "PIO Mode:       %s                %s               %s                 %s\n",
		"?", "?", "?", "?");
	p += sprintf(p, "                %s                     %s\n",
		(reg50 & CFR_INTR_CH0) ? "interrupting" : "polling     ",
		(reg57 & ARTTIM23_INTR_CH1) ? "interrupting" : "polling");
	p += sprintf(p, "                %s                          %s\n",
		(reg71 & 0x04) ? "pending" : "clear  ",
		(reg71 & 0x08) ? "pending" : "clear");
	p += sprintf(p, "                %s                          %s\n",
		(reg71 & 0x10) ? "blocked" : "enabled",
		(reg71 & 0x20) ? "blocked" : "enabled");

	p += sprintf(p, "CFR       = 0x%02x\n", reg50);
	p += sprintf(p, "ARTTIM23  = 0x%02x\n", reg57);
	p += sprintf(p, "MRDMODE   = 0x%02x\n", reg71);

	return p-buffer;	/* => must be less than 4k! */
}
